'****************************************************************************
'*                 GPIOctl V1.0   2024 Darmok                               *
'*        Template code of a program for controlling GPIO board V1.0        *
'*                 Microsoft Basic PDS7.1 (C)                               *
'*                                                                          *
'*    For a description of the input and output values of procedures,       *
'*  as well as explanations, see the code of the corresponding procedures.  *
'*                                                                          *
'*    If the GPIO board controls several devices and several control        *
'*  programs are used, then the distribution of software and hardware       *
'*  resources between them should be organized manually.                    *
'*                                                                          *
'* Notes: 1. Versions of 8250 UART without Scratch Pad Register             *
'*         are not supported by the program.                                *
'*         2. When using the program in a Windows environment, you should   *
'*         disable the corresponding COM port in the system propert         *
'*                                                                          *
'****************************************************************************

DECLARE SUB PortInit ()
DECLARE SUB PutOuts (outs() AS INTEGER)
DECLARE SUB GetOuts (outs() AS INTEGER)
DECLARE SUB GetIns (ins() AS INTEGER)
DECLARE SUB GetAux (bits() AS INTEGER)
DECLARE SUB PutAux (bits() AS INTEGER)
DECLARE FUNCTION GetTRG% ()

DIM SHARED Port AS INTEGER            'required for all procedures
DIM SHARED TRGmode AS INTEGER         'required for the GetTRG function


'****************************************************************************
'-------- Be sure to do this every time you start the program. --------------
'* We check for a hardware reset of the GPIO board and, if necessary,
'  initialize the port.
'* Assign a port address.
   'Port = &H3F8      'COM1 port adress
   Port = &H2F8      'COM2 port adress
   'Port = &H3E8      'COM3 port adress
   'Port = &H2E8      'COM4 port adress
   '*
   PortInit
'*
'------------- Initializing the main variables ------------------------------
'* OUT outputs: outs(#) <-> OUT#   ( SUB PutOuts, SUB GetOuts )
   DIM outs(1 TO 8) AS INTEGER     'any name can be used
   '* We assign all outs(#) the value "do not change". Further
    ' we manipulate only the values of the required outputs.
      FOR outnum% = 1 TO 8: outs(outnum%) = 1: NEXT
   '*
'*
'* IN inputs: IN# -> ins(#)        ( SUB GetIns )
   DIM ins(1 TO 4) AS INTEGER      'any name can be used
'*
'* TRG input: TRG -> GetTRG        ( FUNCTION GetTRG )
   '* We assign the TRG input mode in accordance with the installation
    ' jumper J1 on GPIO board:
    ' J1=OFF - trigger mode, TRGmode=-1;
    ' J1=ON - direct mode, TRGmode= 0.
      TRGmode = 0
   '*
'*
'---------------- Initialize auxiliary variables ----------------------------
'* Auxiliary bits: bits(#)<->BIT#  ( SUB PutAux, SUB GetAUX )
' The auxiliary bits have no effect on the GPIO board, and
' are intended to store any values between calls programs.
    DIM bits(1 TO 5) AS INTEGER    'any name can be used
    '* Set all bits(#) to "do not change". Further we manipulate only
    '  the values of the necessary bits.
       FOR bitnum% = 1 TO 5: bits(bitnum%) = 1: NEXT
    '*
'*
'****************************************************************************

'* Next comes the control program code.
'  ...
'  GetIns ins()
'  ...
'  PutOuts outs()
'  ...
'  GetOuts outs()
'  ...
'  trg% = GetTRG%
'  ...
'  PutAux bits()
'  ...
'  GetAux bits()
'  ...
'*

END

SUB GetAux (bits() AS INTEGER)

'***** The procedure returns the values of the bits stored in the OUT1 bit
'***** of the MCR register and in the RDAI, THREI, RLSI, MSI bits of the
'***** IER register.
'***** There are no input values.
'***** Return values:
'***** if the corresponding bit = 1, then bits(#) = -1;
'***** if the corresponding bit = 0, then bits(#) = 0.
'***** Access to the Port variable is required.

'* Find port addresses.
    aIER% = Port + 1       'Interrupt Enable Register
    aMCR% = Port + 4       'Modem Control Register
'*
'* Read the state of the IER and MCR registers.
    dIER% = INP(aIER%)
    dMCR% = INP(aMCR%)
'*
'* Find the bit value in the MCR and write it to bits(1).
    IF (dMCR% AND 4) = 0 THEN bits(1) = 0 ELSE bits(1) = -1 'from OUT1 bit
'*
'* Find the bit values in IER and put them in bits(#)
    IF (dIER% AND 1) = 0 THEN bits(2) = 0 ELSE bits(2) = -1 'from RDAI bit
    IF (dIER% AND 2) = 0 THEN bits(3) = 0 ELSE bits(3) = -1 'from THREI bit
    IF (dIER% AND 4) = 0 THEN bits(4) = 0 ELSE bits(4) = -1 'from RLSI bit
    IF (dIER% AND 8) = 0 THEN bits(5) = 0 ELSE bits(5) = -1 'from MSI bit
'*

END SUB

SUB GetIns (ins() AS INTEGER)

'***** The procedure returns the values of the current input states.
'***** There are no input values.
'***** Return values:
'***** if IN#=HIGH, then ins(#)=-1;
'***** if IN#=LOW, then ins(#)=0.
'***** Access to the Port variable is required.

'* Find the port address.
   aMSR% = Port + 6      'Modem Status Register
'*
'* Read the state of the inputs from the MSR register.
   dMSR% = INP(aMSR%)
'*
'* Find the corresponding bits and enter their values into ins(#).
   IF (dMSR% AND 128) = 0 THEN ins(1) = 0 ELSE ins(1) = -1    'DCD bit, IN1
   IF (dMSR% AND 32) = 0 THEN ins(2) = 0 ELSE ins(2) = -1     'DSR bit, IN2
   IF (dMSR% AND 16) = 0 THEN ins(3) = 0 ELSE ins(3) = -1     'CTS bit, IN3
   IF (dMSR% AND 64) = 0 THEN ins(4) = 0 ELSE ins(4) = -1     'RI bit,  IN4
'*

END SUB

SUB GetOuts (outs() AS INTEGER)

'***** The procedure returns a copy of the output state values,
'***** saved in the SCR register.
'***** There are no input values.
'***** Return values:
'***** if OUT#=HIGH, then outs(#)=-1;
'***** if OUT#=LOW, then outs(#)=0.
'***** Access to the Port variable is required.

'* Find the port address.
   aSCR% = Port + 7      'Scratch Pad Register
'*
'* Read the saved state of the outputs from the SCR register.
   dSCR% = INP(aSCR%)
'*
'* Find the corresponding bits and enter their values into outs(#).
   IF (dSCR% AND 1) = 0 THEN outs(1) = 0 ELSE outs(1) = -1      'OUT1
   IF (dSCR% AND 2) = 0 THEN outs(2) = 0 ELSE outs(2) = -1      'OUT2
   IF (dSCR% AND 4) = 0 THEN outs(3) = 0 ELSE outs(3) = -1      'OUT3
   IF (dSCR% AND 8) = 0 THEN outs(4) = 0 ELSE outs(4) = -1      'OUT4
   IF (dSCR% AND 16) = 0 THEN outs(5) = 0 ELSE outs(5) = -1     'OUT5
   IF (dSCR% AND 32) = 0 THEN outs(6) = 0 ELSE outs(6) = -1     'OUT6
   IF (dSCR% AND 64) = 0 THEN outs(7) = 0 ELSE outs(7) = -1     'OUT7
   IF (dSCR% AND 128) = 0 THEN outs(8) = 0 ELSE outs(8) = -1    'OUT8
'*

END SUB

FUNCTION GetTRG%

'***** The function returns the trigger state values or the current
'***** TRG input status.
'***** There are no input values.
'***** Return value:
'***** 1.Trigger mode, TRGmode=-1, J1=OFF.
'*****   If since the previous reading there was one
'*****   or several HIGH level pulses with a duration of more than 61
'*****   microseconds (pause between pulses is at least 4.4 microseconds),
'*****   then GetTRG=-1;
'*****   if there have been no pulses since the previous reading
'*****   or they were shorter than 61 microseconds, then GetTRG=0.
'***** 2.Direct mode, TRGmode=0, J1=ON.
'*****   If at the moment of reading there is a LOW level at the TRG input,
'*****   then GetTRG=0;
'*****   if at the moment of reading the TRG input level is HIGH,
'*****   then GetTRG=-1.
'***** Requires access to the Port and TRGmode variables.
'***** Note: The TRG input should be used as either a trigger or as direct.
'*****       The mode is set in hardware using jumper J1 on the GPIO board
'*****       and programmatically using the TRGmode variable.

'* Find port addresses.
    aLSR% = Port + 5      'Line Status Register
    aMCR% = Port + 4      'Modem Control Register
'*
'* Find the trigger state or the current input state.
    SELECT CASE TRGmode
      CASE -1                   'find the trigger state
	IF (INP(aLSR%) AND 16) = 0 THEN
	  GetTRG% = 0           'bit BI=0, there was no pulse
	ELSE
	  GetTRG% = -1          'bit BI=1, pulse was
	END IF
      CASE 0                    'find the current input state
	dLSR% = INP(aLSR%)      'reset the BI bit
	'* Pull input TRG.
	   dMCR% = INP(aMCR%)   'read the MCR register
	   OUT aMCR%, dMCR% AND 254     'bit DTR=0 -> DTR=LOW
	   '* wait 1 mS
	      t# = TIMER + .001
	      WHILE TIMER < t#: WEND
	   '*
	   OUT aMCR%, dMCR% OR 1        'bit DTR=1 -> DTR=HIGH
	   '* wait 1 mS
	      t# = TIMER + .001
	      WHILE TIMER < t#: WEND
	   '*
	'*
	'* Read the current state of the TRG input.
	   IF (INP(aLSR%) AND 16) = 0 THEN
	     GetTRG% = 0         'if bit BI=0, then RXD=LOW
	   ELSE
	     GetTRG% = -1        'if bit BI=1, then RXD=HIGH
	   END IF
	'*
    END SELECT
'*

END FUNCTION

SUB PortInit

'***** Checking for GPIO board hardware reset status (LOW level at inputs
'***** "R" of 74164 and 74273) and initialization port in this case.
'***** There are no return values.
'***** There are no input values.
'***** Access to the Port variable is required.
'***** Note: In PC, setting the OUT2 bit of the MCR register to 1 in hardware
'*****       allows issuing interrupt requests from the COM port. That's why
'*****       we reset it, disabling interruptions.

'* Find port addresses.
    aDLL% = Port          'Divisor Latch LSB
    aDLM% = Port + 1      'Divisor Latch MSB
    aIER% = Port + 1      'Interrupt Enable Register
    aLCR% = Port + 3      'Line Control Register
    aMCR% = Port + 4      'Modem Control Register
    aLSR% = Port + 5      'Line Status Register
    aSCR% = Port + 7      'Scratch Pad Register
'*
'* Check for a hardware reset condition.
    IF (INP(aLCR%) AND 64) = 64 THEN     'bit BC=1, no reset, TXD=HIGH
      EXIT SUB
    ELSE                                 'bit BC=0, reset available, TXD=LOW
      '* Initialize the port.
	 '* Write zeros to the SCR register.
	 '  We bring it into line with the state of the outputs OUT#=LOW.
	    OUT aSCR%, 0
	 '*
	 '* Set the DTR bit and clear other bits in the MCR register.
	 '  Reset auxiliary bit #1, OUT2=0 -> disable interrupts.
	    OUT aMCR%, 1              'bit DTR=1 -> DTR=HIGH
	 '*
	 '* Set the response delay RXD=HIGH.
	    dLCR% = 128               'set format=1s,5,n,1 -> total 7 bits
	    OUT aLCR%, dLCR%          'bit DLAB=1
	    dDLL% = 1: dDLM% = 0      'set 115200 boud, delay=0.061mS
	    OUT aDLL%, dDLL%: OUT aDLM%, dDLM%
	 '*
	 '* Write zeros to the IER register.
	 '  Reset auxiliary bits #2, #3, #4 and #5.
	    dLCR% = 0: OUT aLCR%, dLCR%   'bit DLAB=0
	    dIER% = 0: OUT aIER%, dIER%
	 '*
	 '* Reset the BI bit (reset the trigger at the TRG input).
	    dLSR% = INP(aLSR%)
	 '*
      '*
      '* We remove the hardware reset.
	 '* Set WR(TXD)=HIGH.
	    dLCR% = 64: OUT aLCR%, dLCR%      'bit BC=1 -> TXD=HIGH
	 '*
	 '* We are waiting for 5 mS to remove LOW from inputs "R" 74164
	 '  and 74273.
	    t# = TIMER + .005
	    WHILE TIMER < t#: WEND
	 '*
      '*
    END IF
'*

END SUB

SUB PutAux (bits() AS INTEGER)

'***** The procedure stores the bits(1) values in the OUT1 bit of
'***** the MCR register and bits(2)...bits(5) values in RDAI, THREI, RLSI,
'***** MSI bits of the IER register. Interrupts are not allowed.
'***** There are no return values.
'***** Input values:
'***** if bits(#)=-1, then 1 is stored in the corresponding bit;
'***** if bits(#)= 0, then 0 is stored in the corresponding bit;
'***** if bits(#)= any other number, then the stored value
'***** does not change.
'***** Access to the Port variable is required.

'* Find port addresses.
    aIER% = Port + 1        'Interrupt Enable Register
    aMCR% = Port + 4        'Modem Control Register
'*
'* Read the state of the IER and MCR registers.
    dIER% = INP(aIER%)
    dMCR% = INP(aMCR%)
'*
'* Store the value of bits(1) in the OUT1 bit of the MCR register.
    SELECT CASE bits(1)
      CASE -1
	dMCR% = dMCR% OR 4
	OUT aMCR%, dMCR%
      CASE 0
	dMCR% = dMCR% AND 251
	OUT aMCR%, dMCR%
    END SELECT
'*
'* Store the values bits(2)...bits(5) in bits
'  RDAI, THREI, RLSI, MSI of the IER register.
    bitnum% = 1
    FOR auxbit% = 2 TO 5
      SELECT CASE bits(auxbit%)
	CASE -1
	  dIER% = dIER% OR bitnum%
	CASE 0
	  dIER% = dIER% AND (255 - bitnum%)
      END SELECT
      bitnum% = bitnum% * 2
    NEXT
    OUT aIER%, dIER%
'*

END SUB

SUB PutOuts (outs() AS INTEGER)

'***** Procedure for writing new output states to register 74273
'***** and their copys to the SCR register.
'***** There are no return values.
'***** Input values:
'***** if outs(#) =-1, then OUT#=HIGH;
'***** if outs(#) = 0, then OUT#=LOW;
'***** if outs(#) = any other number, then the state of OUT# does not change.
'***** Access to the Port variable is required.

'* Find port addresses.
    aLCR% = Port + 3        'Line Control Register
    aMCR% = Port + 4        'Modem Control Register
    aSCR% = Port + 7        'Scratch Pad Register
'*
'* Determine the order in which new output states are written to 74164,
'  according to the GPIO board electrical diagram.
    DIM ord(1 TO 8) AS INTEGER
    ord(1) = 1: ord(2) = 2: ord(3) = 3: ord(4) = 4
    ord(5) = 8: ord(6) = 7: ord(7) = 6: ord(8) = 5
'*
'* Assign the initial value of the new state of the outputs.
    newouts% = 0
'*
'* Read the state of the MCR and SCR registers.
    dMCR% = INP(aMCR%)
    oldouts% = INP(aSCR%)
'*
'* Sequential recording of new output states to 74164.
    FOR outnum% = 1 TO 8
      '* Find the bit value to write.
	 '* Find the bit position code.
	    bitcd% = 1
	    FOR order% = 1 TO ord(outnum%) - 1
	      bitcd% = bitcd% * 2
	    NEXT
	 '*
	 SELECT CASE outs(ord(outnum%))
	   CASE 0
	     outbit% = 0
	   CASE -1
	     outbit% = 1
	   CASE ELSE
	     IF (oldouts% AND bitcd%) = 0 THEN
	       outbit% = 0
	     ELSE
	       outbit% = 1
	     END IF
	 END SELECT
      '*
      '* Write the bit to 74164 and write the bit value to newouts%.
	 IF outbit% = 1 THEN
	   newouts% = newouts% OR bitcd%
	   dMCR% = dMCR% OR 2          'bit RTS=1 -> RTS=HIGH
	 ELSE
	   dMCR% = dMCR% AND 253       'bit RTS=0 -> RTS=LOW
	 END IF
	 OUT aMCR%, dMCR% AND 254      'bit DTR=0 -> DTR=LOW
	 OUT aMCR%, dMCR% OR 1         'bit DTR=1 -> DTR=HIGH
      '*
    NEXT
'*
'* Save the new state of the outputs in the SCR register.
    OUT aSCR%, newouts%
'*
'* Rewrite the new state of outputs from 74164 to 74273.
    OUT aLCR%, 0        'bit BC=0 -> TXD=LOW
    OUT aLCR%, 64       'bit BC=1 -> TXD=HIGH
'*

END SUB

